from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="graph_onedrive",
    install_requires=["requests"],  # To-do: specify a min version based on testing
)
